//
// 2up.jit.mesh~.c
// 2D waveguide mesh. 
// by Randy Jones rej@2uptech.com.
// 
// who	when 		what
// rej	11 May 04	created
// rej	June 05		AltiVec optimizations
// rej	Aug 05		choice of termination from negative reflecion to anechoic to positive reflection 
// rej	Apr 07		UB version

#include "jit.common.h"
#include "z_dsp.h"
//#pragma altivec_model on	// needed after z_dsp.h!

static float TENSION_TO_FREQ = 12000.;

//#include <altivec.h>


//#include <math.h>


void *mesh_class;


// TO DO:
// get mesh working scalar. OK
// add mask with zeroes on padded edge. OK
// set mask. OK
// set dim. resize OK
// set freq. OK
// handle denorms. OK
// get working altivec. OK
// add input modes: drum, add. OK
// add audio in. OK
// power normalization. 
// try readout into buffer as in Cook, Essl, O'brien paper. various simplifications (1D, sum towards center)

// ideal 2D kernel from Savioja paper - 1.
// static float s_kernel[9] = {0.09398, 0.312, 0.09398, 0.312, 0.3759, 0.312, 0.09398, 0.312, 0.09398}; // 2. - 1.62392 = 0.37608, savioja center = 0.3759

// no-diagonals kernel could be optimized more but sounds unacceptably bad.
// static float s_kernel[9] = {0., 0.5, 0., 0.5, 0., 0.5, 0., 0.5, 0.0}; 

// mesh[t+1] = mesh[t]Xkernel - mesh[t-1] * (mask * damping const)


typedef struct mesh
{
    t_pxobject x_obj;
	void					*obex;
    long 					sig_in_connected;		// is input signal connected?
    long 					tension_connected;		// is tension signal connected?
	long					interpolate;
    long					lock;
	t_float 				fs;
	t_float 				one_over_fs;
	t_float 				damping;
	t_float 				tension;
	t_float 				prev_tension;
	t_float					reflectivity;
	t_float					cubic;

	t_jit_object *			matrix_out;
	t_jit_object *			matrix_out2;
	t_jit_object *			mesh1;
	t_symbol *				mesh1_name;	
	t_float *				mesh1_data;
	t_jit_object *			mesh2;
	t_symbol *				mesh2_name;	
	t_float *				mesh2_data;
	t_jit_object *			mask;				// mask multiplier for simulation: image * damping
	t_float *				mask_data;
	t_jit_object *			mask_geom;			// input or default geometry for mask
	t_float *				mask_geom_data;
	t_jit_object *			in1_matrix;
	t_float *				in1_matrix_data;
	t_jit_object *			in2_matrix;
	t_float *				in2_matrix_data;
	t_jit_object *			border_matrix;
	t_float *				border_matrix_data;
	t_jit_object *			border1_matrix;
	t_float *				border1_matrix_data;
	t_jit_object *			border2_matrix;
	t_float *				border2_matrix_data;
	
	// input matrices
	t_jit_object *			x0_matrix;
	t_float *				x0_matrix_data;
	t_jit_object *			x1_matrix;
	t_float *				x1_matrix_data;
	t_jit_object *			x2_matrix;
	t_float *				x2_matrix_data;

	t_jit_object *			y0_matrix;
	t_symbol *				y0_matrix_name;	

	t_float *				y0_matrix_data;
	
	t_jit_object *			y1_matrix;
	t_float *				y1_matrix_data;
	t_jit_object *			y2_matrix;
	t_float *				y2_matrix_data;
	
	t_jit_object *			dy_matrix;
	t_symbol *				dy_matrix_name;	
	t_float *				dy_matrix_data;
	
	t_float *				p_in_x0;
	t_float *				p_in_x1;
	t_float *				p_in_x2;
	
	t_float *				p_in_y0;
	t_float *				p_in_y1;
	t_float *				p_in_y2;
	


	int						target_mesh;
	int 					input_pending;
	float					frac_kernel[9];
	
	float clip;
	
	float diffuse;
	float diffuse_inc;
	float theta;

	double					bq_a0;				// biquad coefficients for corners
	double					bq_a1;
	double					bq_a2;
	double					bq_b1;
	double					bq_b2;
	double					xm1[4];				// 4x biquad memory for corners
	double					xm2[4];
	double					ym1[4];				
	double					ym2[4];
	double					delay[4];

	long					dim[2];
	float					driveru[2];
	long					driverpos[2];
	float					driverpos_frac[2];
	float					pickupu[2];
	long					pickuppos[2];

	long					rowbytes;			// same for all internal matrices
	int						variation;			// testing
	long					noisectr;
	unsigned long randlast;
	float accum ;//DEBUG
	long in_perform; // test
	int						needs_resize;
} 	t_mesh;

void *mesh_new(Symbol *s, int ac, Atom *av);
void free_matrices(t_mesh *x);
t_jit_err make_matrices(t_mesh *x);
void mesh_free(t_mesh *x);
t_jit_err mesh_resize(t_mesh *x);
void mesh_do_resize(t_mesh *x, t_symbol *s, long ac, t_atom *av);	
void mesh_assist(t_mesh *x, void *b, long m, long a, char *s);
void mesh_dsp(t_mesh *x, t_signal **sp, short *count);
void mesh_biquad_list(t_mesh *x, t_symbol *s, short argc, t_atom *argv);
void mesh_jit_matrix(t_mesh *x, t_symbol *s, int argc, t_atom *argv);
void mesh_mask(t_mesh *x, t_symbol *s, int argc, t_atom *argv);
void mesh_clear(t_mesh *x);
void mesh_bang(t_mesh *x);
void mesh_step(t_mesh *x);

void mesh_matrix_out(t_mesh *x);
void mesh_matrix_clear(t_mesh *x, t_jit_object *m);
void mesh_set_default_mask(t_mesh *x);
void mesh_set_mask_border(t_mesh *x);
void mesh_set_dim(t_mesh *x, void *attr, long argc, t_atom *argv);
t_jit_err mesh_set_driverpos(t_mesh *x, void *attr, long argc, t_atom *argv);
t_jit_err mesh_set_pickuppos(t_mesh *x, void *attr, long argc, t_atom *argv);
t_jit_err mesh_set_damping(t_mesh *x, void *attr, long argc, t_atom *argv);
t_jit_err mesh_set_tension(t_mesh *x, void *attr, long argc, t_atom *argv);
void mesh_mult_damp_geom(t_mesh *x);

t_int *mesh_perform(t_int *w);
inline void run_mesh(t_mesh *x, t_float in1, t_float t, t_float *out1, t_float *out2);
void make_frac_kernel(t_mesh *x);
void add_input_to_mesh(t_mesh *x);

//inline void run_mesh_step_vector(t_mesh *x, float *p_in, float *p_out, float *p_mask, float *p_kernel);
//inline void run_mesh_step_vector_cylinder(t_mesh *x, float *p_in, float *p_out, float *p_mask, float *p_kernel);
inline void run_mesh_step_scalar(t_mesh *x, float *p_in, float *p_out, float *p_mask, float *p_kernel);

void setmem(void *ss, long n, short b);

//void TurnJavaModeOff( vector unsigned long *oldJavaMode );
//void RestoreJavaMode( vector unsigned long *oldJavaMode );

void main(void)
{
	long attrflags;
	void *classex,*attr;

    setup((struct messlist **)&mesh_class, (method)mesh_new, (method)mesh_free, (short)sizeof(t_mesh), 0L, A_GIMME, 0);

	// add dsp
    dsp_initclass();
    addmess((method)mesh_dsp, "dsp", A_CANT, 0);
	addmess((method)mesh_biquad_list, "list", A_GIMME, 0);	// biquad coeffs in inlet 1
	
	// add attributes
	classex = max_jit_classex_setup(calcoffset(t_mesh,obex));
 	attrflags = JIT_ATTR_GET_DEFER_LOW | JIT_ATTR_SET_USURP_LOW ;

	attr = jit_object_new(_jit_sym_jit_attr_offset_array,"dim",_jit_sym_long,2,attrflags,
		(method)0L,(method)mesh_set_dim, 0, calcoffset(t_mesh, dim));
//	jit_attr_addfilterset_clip(attr,2,0,TRUE,FALSE);
	max_jit_classex_addattr(classex,attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset_array, "driverpos", _jit_sym_float32, 2, attrflags,
		(method)0L, (method)mesh_set_driverpos, calcoffset(t_mesh, driverpos));	
	max_jit_classex_addattr(classex, attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset_array, "pickuppos", _jit_sym_float32, 2, attrflags,
		(method)0L, (method)mesh_set_pickuppos, calcoffset(t_mesh, pickuppos));	
	max_jit_classex_addattr(classex, attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset, "damping", _jit_sym_float32, attrflags,
		(method)0L, (method)mesh_set_damping, calcoffset(t_mesh, damping));	
	max_jit_classex_addattr(classex, attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset, "clip", _jit_sym_float32, attrflags,
		(method)0L, (method)0L, calcoffset(t_mesh, clip));	
	max_jit_classex_addattr(classex, attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset, "diffuse", _jit_sym_float32, attrflags,
		(method)0L, (method)0L, calcoffset(t_mesh, diffuse));	
	max_jit_classex_addattr(classex, attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset, "diffuse_inc", _jit_sym_float32, attrflags,
		(method)0L, (method)0L, calcoffset(t_mesh, diffuse_inc));	
	max_jit_classex_addattr(classex, attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset, "tension", _jit_sym_float32, attrflags,
		(method)0L, (method)mesh_set_tension, calcoffset(t_mesh, tension));	
	max_jit_classex_addattr(classex, attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset, "reflectivity", _jit_sym_float32, attrflags,
		(method)0L, (method)0L, calcoffset(t_mesh, reflectivity));	
	max_jit_classex_addattr(classex, attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset, "cubic", _jit_sym_float32, attrflags,
		(method)0L, (method)0L, calcoffset(t_mesh, cubic));	
	max_jit_classex_addattr(classex, attr);	

	attr = jit_object_new(_jit_sym_jit_attr_offset, "interpolate", _jit_sym_long, attrflags,
		(method)0L, (method)0L, calcoffset(t_mesh, interpolate));	
	max_jit_classex_addattr(classex, attr);	

 	max_jit_classex_standard_wrap(classex,NULL,0);
 		
 	// add other messages
    addmess((method)mesh_assist,"assist",A_CANT,0);
	max_addmethod_usurp_low((method)mesh_jit_matrix, "jit_matrix");
	max_addmethod_usurp_low((method)mesh_mask, "mask");
    addmess((method)mesh_clear,"clear",0);
    addmess((method)mesh_step,"step",0);
    addbang((method)mesh_bang);
}


void *mesh_new(Symbol *s, int argc, Atom *argv)
{
	t_mesh * x = (t_mesh *)max_jit_obex_new(mesh_class, NULL); 	// only max object, no jit object

	// two signal inlets
   	dsp_setup((t_pxobject *)x,2);       	
	max_jit_obex_dumpout_set(x, outlet_new(x,0L)); 		//general purpose outlet(rightmost)
    x->matrix_out2 = outlet_new(x, 0L);
    x->matrix_out = outlet_new(x, 0L);
    
    // stereo signal outlets.
    outlet_new((t_pxobject *)x, "signal");
    outlet_new((t_pxobject *)x, "signal");    
    
	post ("mesh_new.");
   
    // defaults
    x->lock = 1;
	x->interpolate = 1;
    
    x->mesh1 = 0;
    x->mesh1_name = 0;
    x->mesh1_data = 0;
    x->mesh2 = 0;
    x->mesh2_name = 0;
    x->mesh2_data = 0;
    x->mask = 0;
    x->mask_data = 0;
    x->mask_geom = 0;
    x->mask_geom_data = 0;
    x->in1_matrix = 0;
    x->in1_matrix_data = 0;
    x->in2_matrix = 0;
    x->in2_matrix_data = 0;
   	x->border_matrix = 0;
    x->border_matrix_data = 0;
   	x->border1_matrix = 0;
    x->border1_matrix_data = 0;
   	x->border2_matrix = 0;
    x->border2_matrix_data = 0;

	// input matrices
	x->x0_matrix = 0;
	x->x0_matrix_data = 0;
	x->x1_matrix = 0;
	x->x1_matrix_data = 0;
	x->x2_matrix = 0;
	x->x2_matrix_data = 0;

	x->y0_matrix = 0;
	x->y0_matrix_data = 0;
	x->y1_matrix = 0;
	x->y1_matrix_data = 0;
	x->y2_matrix = 0;
	x->y2_matrix_data = 0;
	
	x->dy_matrix = 0;
	x->dy_matrix_data = 0;

    x->dim[0] = 16;
    x->dim[1] = 16;
    
    x->driverpos[0] = 8;
    x->driverpos[1] = 8;
    x->driverpos_frac[0] = 0.;
    x->driverpos_frac[1] = 0.;
    
    x->pickuppos[0] = 8;
    x->pickuppos[1] = 8;
    
    x->damping = 1.;
    x->target_mesh = 0;
	x->input_pending = 0; 
	x->tension = 0.5;
	x->prev_tension = 0.;
	x->reflectivity = 0.;
	x->cubic = 0.1;
    x->noisectr = 0;
    
    x->randlast = jit_rand();
    
    x->accum = 0.;
    
    x->diffuse = 0.;
    x->theta = 0.;
    
    // biquad params
    x->bq_a0 = 0.;
    x->bq_a1 = 0.;
    x->bq_a2 = 0.;
    x->bq_b1 = 0.;
    x->bq_b2 = 0.;
    
    x->clip = 1.;
	
 	max_jit_attr_args(x,argc,argv);
	x->needs_resize = TRUE;
	x->lock = 1;
    
    return (x);
}

void free_matrices(t_mesh *x)
{
	if (x->mesh1)
	{	
		jit_object_free(x->mesh1); 
		x->mesh1 = 0;
		x->mesh1_data = 0;
		x->mesh1_name = 0;
	} 
	if (x->mesh2)
	{
		jit_object_free(x->mesh2); 
		x->mesh2 = 0;
		x->mesh2_data = 0;
		x->mesh2_name = 0;
	}
	if (x->mask)
	{
		jit_object_free(x->mask); 
		x->mask = 0;
		x->mask_data = 0;
	}
	if (x->mask_geom)
	{
		jit_object_free(x->mask_geom); 
		x->mask_geom = 0;
		x->mask_geom_data = 0;
	}
	if (x->in1_matrix)
	{
		jit_object_free(x->in1_matrix); 
		x->in1_matrix = 0;
		x->in1_matrix_data = 0;
	}
	if (x->in2_matrix)
	{
		jit_object_free(x->in2_matrix); 
		x->in2_matrix = 0;
		x->in2_matrix_data = 0;
	}
	if (x->border_matrix)
	{
		jit_object_free(x->border_matrix); 
		x->border_matrix = 0;
		x->border_matrix_data = 0;
	}
	if (x->border1_matrix)
	{
		jit_object_free(x->border1_matrix); 
		x->border1_matrix = 0;
		x->border1_matrix_data = 0;
	}
	if (x->border2_matrix)
	{
		jit_object_free(x->border2_matrix); 
		x->border2_matrix = 0;
		x->border2_matrix_data = 0;
	}
	if (x->x0_matrix)
	{
		jit_object_free(x->x0_matrix); 
		x->x0_matrix = 0;
		x->x0_matrix_data = 0;
	}
	if (x->x1_matrix)
	{
		jit_object_free(x->x1_matrix); 
		x->x1_matrix = 0;
		x->x1_matrix_data = 0;
	}
	if (x->x2_matrix)
	{
		jit_object_free(x->x0_matrix); 
		x->x0_matrix = 0;
		x->x0_matrix_data = 0;
	}
	if (x->y0_matrix)
	{
		jit_object_free(x->y0_matrix); 
		x->y0_matrix = 0;
		x->y0_matrix_data = 0;
	}
	if (x->y1_matrix)
	{
		jit_object_free(x->y1_matrix); 
		x->y1_matrix = 0;
		x->y1_matrix_data = 0;
	}
	if (x->y2_matrix)
	{
		jit_object_free(x->y2_matrix); 
		x->y2_matrix = 0;
		x->y2_matrix_data = 0;
	}
	if (x->dy_matrix)
	{
		jit_object_free(x->dy_matrix); 
		x->dy_matrix = 0;
		x->dy_matrix_data = 0;
	}
}


t_jit_err make_matrices(t_mesh *x)
{
	t_jit_err err = JIT_ERR_NONE;
	t_jit_matrix_info info;

	jit_matrix_info_default(&info);
	info.type = _jit_sym_float32;
	info.dimcount = 2;
	info.planecount = 1;
	info.dim[0] = x->dim[0];		
	info.dim[1] = x->dim[1];
	
	// mesh 1	
	if (!(x->mesh1 = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	x->mesh1_name = jit_symbol_unique();

	if (!(x->mesh1 = jit_object_method(x->mesh1, _jit_sym_register, x->mesh1_name)))
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	jit_object_method(x->mesh1, _jit_sym_getdata, &x->mesh1_data);	
	
	// get rowbytes
	jit_object_method(x->mesh1,_jit_sym_getinfo,&info);
	x->rowbytes = info.dimstride[1];

	// mesh 2
	if (!(x->mesh2 = jit_object_new(gensym("jit_matrix"), &info)))
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	x->mesh2_name = jit_symbol_unique();
	if (!(x->mesh2 = jit_object_method(x->mesh2, _jit_sym_register, x->mesh2_name)))
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	jit_object_method(x->mesh2, _jit_sym_getdata, &x->mesh2_data);
	
	// mask
	if (!(x->mask = jit_object_new(gensym("jit_matrix"), &info)))
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	if (!(x->mask = jit_object_method(x->mask, _jit_sym_register, jit_symbol_unique()))) 
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	jit_object_method(x->mask, _jit_sym_getdata, &x->mask_data);
	
	// mask geom
	if (!(x->mask_geom = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	if (!(x->mask_geom = jit_object_method(x->mask_geom, _jit_sym_register, jit_symbol_unique()))) 
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	jit_object_method(x->mask_geom, _jit_sym_getdata, &x->mask_geom_data);
	
	// input matrix
	if (!(x->in1_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	if (!(x->in1_matrix = jit_object_method(x->in1_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	jit_object_method(x->in1_matrix, _jit_sym_getdata, &x->in1_matrix_data);
	
	// input matrix 2
	if (!(x->in2_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	if (!(x->in2_matrix = jit_object_method(x->in2_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	jit_object_method(x->in2_matrix, _jit_sym_getdata, &x->in2_matrix_data);

	// border matrix
	if (!(x->border_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	if (!(x->border_matrix = jit_object_method(x->border_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	jit_object_method(x->border_matrix, _jit_sym_getdata, &x->border_matrix_data);

	// border matrix 1
	if (!(x->border1_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	if (!(x->border1_matrix = jit_object_method(x->border1_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	jit_object_method(x->border1_matrix, _jit_sym_getdata, &x->border1_matrix_data);

	// border matrix 2
	if (!(x->border2_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	if (!(x->border2_matrix = jit_object_method(x->border2_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM;
		goto bail;
	}
	jit_object_method(x->border2_matrix, _jit_sym_getdata, &x->border2_matrix_data);



	// input FIR and diff. matrices
	
	// x0
	if (!(x->x0_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	if (!(x->x0_matrix = jit_object_method(x->x0_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	jit_object_method(x->x0_matrix, _jit_sym_getdata, &x->x0_matrix_data);

	// x1
	if (!(x->x1_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	if (!(x->x1_matrix = jit_object_method(x->x1_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	jit_object_method(x->x1_matrix, _jit_sym_getdata, &x->x1_matrix_data);

	// x2
	if (!(x->x2_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	if (!(x->x2_matrix = jit_object_method(x->x2_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	jit_object_method(x->x2_matrix, _jit_sym_getdata, &x->x2_matrix_data);
	
	// y0
	if (!(x->y0_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	if (!(x->y0_matrix = jit_object_method(x->y0_matrix, _jit_sym_register, x->y0_matrix_name = jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	jit_object_method(x->y0_matrix, _jit_sym_getdata, &x->y0_matrix_data);

	// y1
	if (!(x->y1_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	if (!(x->y1_matrix = jit_object_method(x->y1_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	jit_object_method(x->y1_matrix, _jit_sym_getdata, &x->y1_matrix_data);

	// y2
	if (!(x->y2_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	if (!(x->y2_matrix = jit_object_method(x->y2_matrix, _jit_sym_register, jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	jit_object_method(x->y2_matrix, _jit_sym_getdata, &x->y2_matrix_data);

	// dy
	if (!(x->dy_matrix = jit_object_new(gensym("jit_matrix"), &info))) 
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	if (!(x->dy_matrix = jit_object_method(x->dy_matrix, _jit_sym_register, x->dy_matrix_name = jit_symbol_unique())))
	{
		err = JIT_ERR_OUT_OF_MEM; goto bail;
	}
	jit_object_method(x->dy_matrix, _jit_sym_getdata, &x->dy_matrix_data);

bail:
	return err;
}


void mesh_free(t_mesh *x)
{
	x->lock = 1;
	free_matrices(x);

	//only max object, no jit object
	max_jit_obex_free(x);
	dsp_free((t_pxobject *)x);
}


/*void jit_tactex_bang(t_jit_tactex *x)
{
	defer(x, (method)serial_read, NULL, 0, NULL);
}

void serial_read(t_jit_tactex *x, t_symbol *s, long ac, t_atom *av);	

*/


t_jit_err mesh_resize(t_mesh *x)
{
	x->needs_resize = TRUE;
	return JIT_ERR_NONE;
}


void mesh_do_resize(t_mesh *x, t_symbol *s, long ac, t_atom *av)
{
	t_jit_matrix_info info;
	t_jit_err err = JIT_ERR_NONE;
	t_atom a[2];

	free_matrices(x);
	err = make_matrices(x);
		
	if (err == JIT_ERR_NONE)
	{	
   
		// set pickup / driver positions based on new dims
		jit_atom_setfloat(&a[0], x->driveru[0]);
		jit_atom_setfloat(&a[1], x->driveru[1]);
		mesh_set_driverpos(x, 0, 2, a);
		
		jit_atom_setfloat(&a[0], x->pickupu[0]);
		jit_atom_setfloat(&a[1], x->pickupu[1]);
		mesh_set_pickuppos(x, 0, 2, a);
	
		jit_object_method(x->mesh1,_jit_sym_getinfo,&info);
		x->rowbytes = info.dimstride[1];

		x->lock = 0;
		
		mesh_clear(x);		
		mesh_set_default_mask(x);
		mesh_mult_damp_geom(x);	
		
	}
	else
	{
		post("2up.jit.mesh~: couldn't resize!");
	}
}


void mesh_assist(t_mesh *x, void *b, long msg, long arg, char *dst)
{
	if (msg==1) 
	{
		if (arg == 0)
		{
			// input 1
			sprintf(dst, "driver signal, jitter matrix, or bang in");
		}
		else if (arg == 1)
		{
			sprintf(dst, "tension signal input");
		}
	} 
	else if (msg==2) 
	{
		if (arg == 0)
		{
			sprintf(dst, "signal out L");
		}
		else if (arg == 1)
		{
			sprintf(dst, "signal out R");
		}
		else if (arg == 2)
		{
			sprintf(dst, "matrix out");
		}
		else if (arg == 3)
		{
			sprintf(dst, "dumpout");
		}
	}
}


void mesh_dsp(t_mesh *x, t_signal **sp, short *count)
{
	x->fs = sp[0]->s_sr;
	x->one_over_fs = 1.0/x->fs;
	x->sig_in_connected = count[0];
	x->tension_connected = count[1];
	
	dsp_add(mesh_perform, 6, x, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec, sp[3]->s_vec, sp[0]->s_n); // obj, in, tension, out1, out2, num_samples
}

void mesh_biquad_list(t_mesh *x, t_symbol *s, short argc, t_atom *argv)
{
	short i;
	double a0 = x->bq_a0, a1 = x->bq_a1, a2 = x->bq_a2, b1 = x->bq_b1, b2 = x->bq_b2;
	
	for (i=0; i < argc; i++) 
	{
		switch (i) 
		{
			case 0: a0 = atom_getfloatarg(i,argc,argv); break;
			case 1: a1 = atom_getfloatarg(i,argc,argv); break;
			case 2: a2 = atom_getfloatarg(i,argc,argv); break;
			case 3: b1 = atom_getfloatarg(i,argc,argv); break;
			case 4: b2 = atom_getfloatarg(i,argc,argv); break;
		}
	}
	x->bq_a0 = a0;
	x->bq_a1 = a1;
	x->bq_a2 = a2;
	x->bq_b1 = b1;
	x->bq_b2 = b2; 
	post ("got biquad list: %f %f %f %f %f\n", a0, a1, a2, b1, b2);
}



void mesh_jit_matrix(t_mesh *x, t_symbol *s, int argc, t_atom *argv)
{
	t_jit_matrix_info 		info;
	t_symbol * 				matrix_name = 0;
	void *					matrix = 0;
	void *					p_data = 0;
	long					rowbytes;
	long					height, width;
	t_matrix_conv_info 		mcinfo = {0};
	
	// get matrix
	matrix_name = jit_atom_getsym(&argv[0]);
	if (matrix_name != _jit_sym_nothing) 
	{
		matrix = jit_object_findregistered(matrix_name);
	}
	if (!matrix) 
	{
		error ("2up.jit.mesh~: couldn't get matrix object %s!", matrix_name->s_name);
		return;
	}
	
	jit_object_method(matrix,_jit_sym_getinfo,&info);
	jit_object_method(matrix,_jit_sym_getdata,&p_data);
	if (p_data == 0)
	{
		error("2up.jit.mesh~: null data ptr for matrix!");
		return;
	}	
	if (info.dimcount != 2)
	{
		error("2up.jit.mesh~: input matrix must be 2D.");
		return;
	}	
	if (info.planecount != 1)
	{
		error("2up.jit.mesh~: input matrix must be 1 plane.");
		return;
	}	
	if (info.type != _jit_sym_float32)
	{
		error("2up.jit.mesh~: sorry, float32 matrix needed.");
		return;
	}
	height = info.dim[1];
	width = info.dim[0];
	rowbytes = info.dimstride[1];
	
	// copy input to internal input matrix
	{
		mcinfo.flags |= JIT_MATRIX_CONVERT_INTERP;
		jit_object_method(x->in1_matrix, _jit_sym_frommatrix, matrix, &mcinfo);
	}
	
	add_input_to_mesh(x);
	
//	x->input_pending = 1;
}

void mesh_mask(t_mesh *x, t_symbol *s, int argc, t_atom *argv)
{
	t_jit_matrix_info 		info;
	t_symbol * 				matrix_name = 0;
	void *					matrix = 0;
	void *					p_data = 0;
	long					rowbytes;
	long					height, width;
	t_matrix_conv_info 		mcinfo = {0};

	// check for "jit_matrix"
	matrix_name = jit_atom_getsym(&argv[0]);
	if (matrix_name != _jit_sym_jit_matrix)
	{
		error ("2up.jit.mesh~: mask must be a matrix.");
		return;
	}
	// get matrix
	matrix_name = jit_atom_getsym(&argv[1]);
	if (matrix_name != _jit_sym_nothing) 
	{
		matrix = jit_object_findregistered(matrix_name);
	}
	if (!matrix) 
	{
		error ("2up.jit.mesh~: couldn't get matrix object %s!", matrix_name->s_name);
		return;
	}
	
	jit_object_method(matrix,_jit_sym_getinfo,&info);
	jit_object_method(matrix,_jit_sym_getdata,&p_data);
	if (p_data == 0)
	{
		error("2up.jit.mesh~: null data ptr for matrix!");
		return;
	}	
	if (info.dimcount != 2)
	{
		error("2up.jit.mesh~: input matrix must be 2D.");
		return;
	}	
	if (info.planecount != 1)
	{
		error("2up.jit.mesh~: input matrix must be 1 plane.");
		return;
	}	
	if (info.type != _jit_sym_float32)
	{
		error("2up.jit.mesh~: sorry, float32 matrix needed.");
		return;
	}
	height = info.dim[1];
	width = info.dim[0];
	rowbytes = info.dimstride[1];
	
	// copy input to mask
	{
		mcinfo.flags |= JIT_MATRIX_CONVERT_INTERP;
		jit_object_method(x->mask_geom, _jit_sym_frommatrix, matrix, &mcinfo);
	}
	
	// multiply by damping
	mesh_set_mask_border(x);

	mesh_mult_damp_geom(x);
}


void mesh_clear(t_mesh *x)
{
	int i;

	mesh_matrix_clear(x, x->mesh1);
	mesh_matrix_clear(x, x->mesh2);
	mesh_matrix_clear(x, x->in1_matrix);
	mesh_matrix_clear(x, x->border_matrix);
	mesh_matrix_clear(x, x->border1_matrix);
	mesh_matrix_clear(x, x->border2_matrix);

	mesh_matrix_clear(x, x->x0_matrix);
	mesh_matrix_clear(x, x->x1_matrix);
	mesh_matrix_clear(x, x->x2_matrix);

	mesh_matrix_clear(x, x->y0_matrix);
	mesh_matrix_clear(x, x->y1_matrix);
	mesh_matrix_clear(x, x->y2_matrix);
	mesh_matrix_clear(x, x->dy_matrix);

   // biquad history
    for(i=0; i<4; i++)
    {
    	x->xm1[i] = 0.;
    	x->xm2[i] = 0.;
    	x->ym1[i] = 0.;
    	x->ym2[i] = 0.;
    	x->delay[i] = 0.;
    }
    
    x->prev_tension = 0;
}


void mesh_bang(t_mesh *x)
{
	if (x->lock) return;
	mesh_matrix_out(x);
}

// for testing or non-audio work: perform method without signal stuff.
void mesh_step(t_mesh *x)
{
	float *p_inmesh, *p_outmesh;

	if (x->lock) return;

	// swap meshes
	x->target_mesh = (x->target_mesh == 0);
	p_inmesh = x->target_mesh ? x->mesh1_data : x->mesh2_data;
	p_outmesh = x->target_mesh ? x->mesh2_data : x->mesh1_data;

//	run_mesh_step_vector(x, p_inmesh, p_outmesh, x->mask_data, x->frac_kernel);
	run_mesh_step_scalar(x, p_inmesh, p_outmesh, x->mask_data, x->frac_kernel);
}


void mesh_matrix_out(t_mesh *x)
{
	Atom av[1];
	t_symbol * p_name;
	if (x->lock) return;
	
	// write sim matrix
	p_name = (x->target_mesh) ? x->mesh2_name : x->mesh1_name;
	
	jit_atom_setsym(av, p_name);
	outlet_anything(x->matrix_out, _jit_sym_jit_matrix, 1, av);	
	
	
	// TEST
	p_name = x->dy_matrix_name;	
	jit_atom_setsym(av, p_name);
	outlet_anything(x->matrix_out2, _jit_sym_jit_matrix, 1, av);	
	
}

void mesh_matrix_clear(t_mesh *x, t_jit_object *m)
{
	void * pm = 0;
	long bytes = x->dim[1] * x->rowbytes;
	
	if (x->lock) return;
	if (!m) 
	{
		error("jit.2up.mesh~: clearing null matrix!");
		return;
	}
	
	jit_object_method(m, _jit_sym_getdata, &pm);
	if (pm)
	{
		setmem(pm, bytes, 0);
	}
	else
	{
		post ("_mesh_clear: couldn't get matrix data ptr!");
	}
}

// set default mask: all ones except border.
void mesh_set_default_mask(t_mesh *x)
{
	float *p;
	float zero = 0.;
	float one = 1.;
	float f;
	int i, j;	
	
	if (x->lock) return;
	
	for (i=0; i<x->dim[1]; i++)
	{
		p = (float *)((long)x->mask_geom_data + (long)(i*x->rowbytes));
		for (j=0; j<x->dim[0]; j++)
		{
			if ((i==0)||(i==x->dim[1]-1)||(j==0)||(j==x->dim[0]-1))
			{
				f = zero;
			}
			else
			{
				f = one;
			}
			p[j] = f;
		}
	}
}


// set border of incoming mask to 0.
void mesh_set_mask_border(t_mesh *x)
{
	float *p;
	int i, j;	
	
	if (x->lock) return;
	
	for (i=0; i<x->dim[1]; i++)
	{
		p = (float *)((long)x->mask_geom_data + (long)(i*x->rowbytes));
		for (j=0; j<x->dim[0]; j++)
		{
			if ((i==0)||(i==x->dim[1]-1)||(j==0)||(j==x->dim[0]-1))
			{
				p[j] = 0.;
			}
		}
	}
}


void mesh_set_dim(t_mesh *x, void *attr, long argc, t_atom *argv)
{
	t_jit_err err;
	if (x->lock) return;
	
	if (argc&&argv) 
	{
		x->dim[0] = MAX(jit_atom_getlong(&argv[0]), 4);	
		x->dim[1] = MAX(jit_atom_getlong(&argv[1]), 4);	
		
		err = mesh_resize(x);
	}
}

t_jit_err mesh_set_driverpos(t_mesh *x, void *attr, long argc, t_atom *argv)
{
	float px, py;	// position from 0. to 1.
	float dx, dy;
	if (argc&&argv) 
	{
		x->driveru[0] = jit_atom_getfloat(&argv[0]);
		x->driveru[1] = jit_atom_getfloat(&argv[1]);	
		px = CLAMP(x->driveru[0], 0., 1.);	
		py = CLAMP(x->driveru[1], 0., 1.);
		dx = px * (x->dim[0] - 3.) + 1.;
		dy = py * (x->dim[1] - 3.) + 1.;
		// integer position
		x->driverpos[0] = jit_math_floor(dx);
		x->driverpos[1] = jit_math_floor(dy);
		// remainder
		x->driverpos_frac[0] = dx - (float)x->driverpos[0];
		x->driverpos_frac[1] = dy - (float)x->driverpos[1];
	}
	return JIT_ERR_NONE;
}

t_jit_err mesh_set_pickuppos(t_mesh *x, void *attr, long argc, t_atom *argv)
{
	float px, py;	// position from 0. to 1.
	float dx, dy;
	if (argc&&argv) 
	{
		x->pickupu[0] = jit_atom_getfloat(&argv[0]);
		x->pickupu[1] = jit_atom_getfloat(&argv[1]);	
		px = CLAMP(x->pickupu[0], 0., 1.);	
		py = CLAMP(x->pickupu[1], 0., 1.);	
		dx = px * (x->dim[0] - 3.) + 1.;
		dy = py * (x->dim[1] - 3.) + 1.;
		// integer position
		x->pickuppos[0] = jit_math_floor(dx);
		x->pickuppos[1] = jit_math_floor(dy);
	}
	return JIT_ERR_NONE;
}


t_jit_err mesh_set_damping(t_mesh *x, void *attr, long argc, t_atom *argv)
{
	if (argc&&argv) 
	{
		x->damping = jit_atom_getfloat(argv);		
		mesh_mult_damp_geom(x);
	}
	return JIT_ERR_NONE;
}

// multiply mask geom * damping
void mesh_mult_damp_geom(t_mesh *x)
{
	float *p_in, *p_out;
	int i, j;	
	
	if (x->lock) return;

	for (i=0; i<x->dim[1]; i++)
	{
		p_in = (float *)((long)x->mask_geom_data + (long)(i*x->rowbytes));
		p_out = (float *)((long)x->mask_data + (long)(i*x->rowbytes));
		for (j=0; j<x->dim[0]; j++)
		{
			p_out[j] = p_in[j]*x->damping;
		}
	}
}


t_jit_err mesh_set_tension(t_mesh *x, void *attr, long argc, t_atom *argv)
{
	if (argc&&argv) 
	{
		x->tension = MAX(MIN(1.0, jit_atom_getfloat(argv)), 0.);		
		make_frac_kernel(x);
	}
	return JIT_ERR_NONE;
}



// odds 'n' ends

// quick and dirty sin approximation 
inline float sin_e(float x);
inline float sin_e(float x)
{
    return x - (x*x*x*0.1666666666666);
}

// ----------------------------------------------------------------
// DSP routines

t_int *mesh_perform(t_int *w)
{
 	t_mesh *x = (t_mesh *)(w[1]);
   	t_float *p_in = (t_float *)(w[2]);
   	t_float *p_tension = (t_float *)(w[3]);
    t_float *p_out1 = (t_float *)(w[4]);
    t_float *p_out2 = (t_float *)(w[5]);
 	long n = w[6];
	int i=0, j=0, i2 = 0, j2 = 0;
    t_float inscale, outscale, sample;
	double *p_m_excite, *px0, *px1, *px2, *py0, *py1, *py2, *pdy, *p_m_out, *p_m_in;
	double a0 = x->bq_a0, a1 = x->bq_a1, a2 = x->bq_a2, b1 = x->bq_b1, b2 = x->bq_b2;
	
	int width = x->dim[0];
	int height = x->dim[1];

	float *p_inmesh, *p_outmesh;
    float *p_pickup, *p_driver;
	
	// resize if needed
	if (x->needs_resize)
	{
		x->needs_resize = FALSE;
		x->lock = 1;	// lock DSP
		defer_low(x, (method)mesh_do_resize, NULL, 0, NULL); // will unlock DSP when completed
	}

	if (x->lock) goto bail;
	
	if (x->x_obj.z_disabled || x->lock)	// clear vector
	{
		while(n--)
		{					
			p_out1[i] = 0;
			p_out2[i] = 0;
			i++;
		}
	}
	else // run mesh
	{
		while(n--)
		{					
			// rotate input pointers rather than copy data
			// swap meshes
			x->target_mesh = (x->target_mesh == 0);
			p_inmesh = x->target_mesh ? x->mesh1_data : x->mesh2_data;
			p_outmesh = x->target_mesh ? x->mesh2_data : x->mesh1_data;
			
			if (x->tension_connected)
			{
				// calculate tension kernel for next sample
				x->tension = p_tension[i];
				make_frac_kernel(x);
			} 
			inscale = (x->tension);
			outscale = 8.0 * (x->tension);
			
			// bandlimit input with FIR filters, H(z) - 1 - z^(-2)
			
			// filter
			for (i2=0; i2<height; i2++)
			{
				p_m_excite = (float *)((long)x->in2_matrix_data + (long)(i2*x->rowbytes));

				px0 = (float *)((long)x->x0_matrix_data + (long)(i2*x->rowbytes));
				px1 = (float *)((long)x->x1_matrix_data + (long)(i2*x->rowbytes));
				px2 = (float *)((long)x->x2_matrix_data + (long)(i2*x->rowbytes));
				
				py0 = (float *)((long)x->y0_matrix_data + (long)(i2*x->rowbytes));
				py1 = (float *)((long)x->y1_matrix_data + (long)(i2*x->rowbytes));
				pdy = (float *)((long)x->dy_matrix_data + (long)(i2*x->rowbytes));
				p_m_in = (float *)((long)p_inmesh + (long)(i2*x->rowbytes));
				p_m_out = (float *)((long)p_outmesh + (long)(i2*x->rowbytes));
					
				for(j2=0; j2<width; j2++)
				{
					// rotate

					py2[j2] = py1[j2];
					py1[j2] = py0[j2];
					
					px2[j2] = px1[j2];
					px1[j2] = px0[j2];
					px0[j2] = p_m_excite[j2];
					
					// add input
					
					// IIR filter input					
					// IIR filter
					// y[n] = a0 * x[n] + a1 * x[n-1] + a2 * x[n-2] - b1 * y[n-1] - b2 * y[n-2]
					// py0[j2] = a0 * px0[j2] + a1*px1[j2] + a2*px2[j2] + b1*py1[j2] + b2*py2[j2];		
					
					py0[j2] = p_m_excite[j2];
					
					// dy/dt
					pdy[j2] = py0[j2] - py1[j2];
					
					// directional hack. 
			//		if (pdy[j2] < 0.) 
			//		{
			//			pdy[j2] *= 0.;
			//		}
					
					// add diff to mesh, 
					p_m_in[j2] += pdy[j2];
				}
			}
			
			// deinterpolate signal to 4 mesh points
			if (x->sig_in_connected)
			{
				sample = p_in[i] * inscale;
				p_driver = (float *)(((long)p_inmesh) + (x->driverpos[1]*x->rowbytes));
				p_driver[x->driverpos[0]] += sample * (1.0 - x->driverpos_frac[0]) * (1.0 - x->driverpos_frac[1]) ;
				p_driver[x->driverpos[0]+1] += sample * (x->driverpos_frac[0]) * (1.0 - x->driverpos_frac[1]) ;
				p_driver = (float *)(((long)p_inmesh) + ((x->driverpos[1]+1)*x->rowbytes));
				p_driver[x->driverpos[0]] += sample * (1.0 - x->driverpos_frac[0]) * (x->driverpos_frac[1]) ;
				p_driver[x->driverpos[0]+1] += sample * (x->driverpos_frac[0]) * (x->driverpos_frac[1]) ;
			}
			
			// comment out one of these lines for either vector or scalar operation. 
			
			// SSE version does not exist yet. 
		//	run_mesh_step_vector(x, p_inmesh, p_outmesh, x->mask_data, x->frac_kernel);
			run_mesh_step_scalar(x, p_inmesh, p_outmesh, x->mask_data, x->frac_kernel);
			
/*			// filter
			for (i2=0; i2<height; i2++)
			{
				p_m_excite = (float *)((long)x->in2_matrix_data + (long)(i2*x->rowbytes));

				px0 = (float *)((long)x->x0_matrix_data + (long)(i2*x->rowbytes));
				px1 = (float *)((long)x->x1_matrix_data + (long)(i2*x->rowbytes));
				px2 = (float *)((long)x->x2_matrix_data + (long)(i2*x->rowbytes));
				
				py0 = (float *)((long)x->y0_matrix_data + (long)(i2*x->rowbytes));
				py1 = (float *)((long)x->y1_matrix_data + (long)(i2*x->rowbytes));
				pdy = (float *)((long)x->dy_matrix_data + (long)(i2*x->rowbytes));
				p_m_in = (float *)((long)p_inmesh + (long)(i2*x->rowbytes));
				p_m_out = (float *)((long)p_outmesh + (long)(i2*x->rowbytes));
					
				for(j2=0; j2<width; j2++)
				{
					p_m_out[j2] = py0[j2];	// TESTING
				}
			}
*/			

			// get ptr to center of output mesh // PRECALCULATE!			
			p_pickup = (float *)(((long)p_outmesh) + x->pickuppos[1]*x->rowbytes);
			
			// make stereo pair by reflecting pickup location around y axis
			p_out1[i] = outscale * p_pickup[x->pickuppos[0]]; 
			p_out2[i] = outscale * p_pickup[x->dim[0] - 1 - x->pickuppos[0]]; 		

			i++;
		}
	}

	x->in_perform = 0;
bail:	
    return (w+7);
}

/*
void postvec(char *label, vector float * v);
void postvec(char *label, vector float * v)
{
	static floatToVector temp;
	temp.vFlt = *v;
	post ("%s: %f, %f, %f, %f", label, temp.flt[0],temp.flt[1],temp.flt[2],temp.flt[3]);
}
*/


// opt: make sure this is getting inlined.  maybe not possible with altivec (VRSAVE)?

// opt: make sure vecLib is still a lot slower.

// opt: it's time for SIMG4.
/*
inline void run_mesh_step_vector(t_mesh *x, float *p_in, float *p_out, float *p_mask, float *p_kernel)
{
	long i, j;
	long width = x->dim[0];
	long height = x->dim[1];
	float *pr_border, *pr_border1, *pr_border2;
	float *pr_out1, *pr_out2, *pr_out3;
	float *pr_in1, *pr_in2, *pr_in3;
	float r;

	float * prm;
	float * p_k = p_kernel;
	float k, e, c;				// corner, edge, middle
	long half = width << 1;
	int offset = 0;
	long vectors = width << 2;	// width must be a multiple of 4
	long v = 0;
	
	float tsq = x->tension;
	float one_over_tsq = 1.0 / tsq;
	
//	floatToVector ktemp;		// in z_altivec.h

//	vector unsigned long oldJavaMode;
	vector float vzero, vmin, vmax;
	vector float vmask, vout1;
	vector float vout2, vout3, vout4, vprevout;
	vector float t1, t2, t3, t4;
	vector float vcurrtop, vprevtop, vnexttop;
	vector float vcurrmid, vprevmid, vnextmid;
	vector float vcurrbot, vprevbot, vnextbot;
	
	vector float vkaprev, vkacurr1, vkacurr2, vkacurr3, vkacurr4, vkanext;
	vector float vkbprev, vkbcurr1, vkbcurr2, vkbcurr3, vkbcurr4, vkbnext;
	
	// eliminate denormals by turning off Java mode (already off in MSP signal chain)
	// TurnJavaModeOff ( &oldJavaMode );
	
            	
	// store a vector containing coefficients for each possible shift of kernel 
	// within vector, including elements in previous and next vectors.
	// for first row of kernel: [k, e, k] we have
	//
	// | prev       | curr       | next       |
	// | 0, 0, 0, k | e, k, 0, 0 | 0, 0, 0, 0 |
	// | 0, 0, 0, 0 | k, e, k, 0 | 0, 0, 0, 0 |
	// | 0, 0, 0, 0 | 0, k, e, k | 0, 0, 0, 0 |
	// | 0, 0, 0, 0 | 0, 0, k, e | k, 0, 0, 0 |
	//
	// 2 distinct kernel rows * 6 nonzero positions = 12 vectors.
	
	// store masks for four float positions within vector
	
	// make constants
	vzero = vec_xor(vzero, vzero);	
	ktemp.flt[0] = ktemp.flt[1] = ktemp.flt[2] = ktemp.flt[3] = -x->clip * 0.01 * one_over_tsq * one_over_tsq;
	vmin = vec_ld(0, ktemp.flt);
	ktemp.flt[0] = ktemp.flt[1] = ktemp.flt[2] = ktemp.flt[3] = x->clip * 0.01 *  one_over_tsq * one_over_tsq;
	vmax = vec_ld(0, ktemp.flt);
	
	// get kernel coefficients
	k = p_k[0];
	e = p_k[1];
	c = p_k[4];
	
	// make kernel vectors
	// row a
	ktemp.flt[0] = 0.;
	ktemp.flt[1] = k;
	ktemp.flt[2] = e;
	ktemp.flt[3] = k;
	vkacurr3 = vec_ld(0, ktemp.flt);
	vkacurr2 = vec_sld(vkacurr3, vzero, 4);
	vkacurr1 = vec_sld(vkacurr3, vzero, 8);
	vkaprev = vec_sld(vzero, vkacurr3, 8);
	vkacurr4 = vec_sld(vzero, vkacurr3, 12);
	vkanext = vec_sld(vkaprev, vzero, 12);
	// row b
	ktemp.flt[0] = 0.;
	ktemp.flt[1] = e;
	ktemp.flt[2] = c;
	ktemp.flt[3] = e;
	vkbcurr3 = vec_ld(0, ktemp.flt);
	vkbcurr2 = vec_sld(vkbcurr3, vzero, 4);
	vkbcurr1 = vec_sld(vkbcurr3, vzero, 8);
	vkbprev = vec_sld(vzero, vkbcurr3, 8);
	vkbcurr4 = vec_sld(vzero, vkbcurr3, 12);
	vkbnext = vec_sld(vkbprev, vzero, 12);
	

//DEBUG
//post ("kernel thangs:");
//postvec("prev:", &vkaprev);
//postvec("c1:", &vkacurr1);
//postvec("c2:", &vkacurr2);
//postvec("c3:", &vkacurr3);
//postvec("c4:", &vkacurr4);
//postvec("next:", &vkanext);


	for(i=1; i<height-1; i++)
	{
		// row ptrs
		pr_in1 = (float *)((long)p_in + (long)((i-1)*x->rowbytes));
		pr_in2 = (float *)((long)p_in + (long)(i*x->rowbytes));
		pr_in3 = (float *)((long)p_in + (long)((i+1)*x->rowbytes));
		prm = (float *)((long)p_mask + (long)(i*x->rowbytes));
		pr_out2 = (float *)((long)p_out + (long)(i*x->rowbytes));
		
		// setup first vector column
		vcurrtop = vzero;
		vcurrmid = vzero;
		vcurrbot = vzero;
		vnexttop = vec_ld(0, pr_in1);
		vnextmid = vec_ld(0, pr_in2);
		vnextbot = vec_ld(0, pr_in3);
		
		// do vector columns
		for (j=0; j<(width)*4; j+=16)
		{	
			// get input vectors
			vprevtop = vcurrtop;
			vprevmid = vcurrmid;
			vprevbot = vcurrbot;
			vcurrtop = vnexttop;
			vcurrmid = vnextmid;
			vcurrbot = vnextbot;
			
			// last column read will read past matrix, then be multiplied by 0 (mask). 
			// on current mem arch this shouldn't be a problem.
			vnexttop = vec_ld((j + 16), pr_in1);
			vnextmid = vec_ld((j + 16), pr_in2);
			vnextbot = vec_ld((j + 16), pr_in3);			

			vout1 = vec_madd(vprevtop, vkaprev, vzero);
			vout3 = vec_madd(vcurrtop, vkacurr3, vzero);
			vout1 = vec_madd(vcurrtop, vkacurr1, vout1);
			vout1 = vec_madd(vcurrmid, vkbcurr1, vout1);
			vout3 = vec_madd(vcurrmid, vkbcurr3, vout3);
			vout1 = vec_madd(vprevmid, vkbprev, vout1);
			vout1 = vec_madd(vcurrbot, vkacurr1, vout1);
			vout3 = vec_madd(vcurrbot, vkacurr3, vout3);
			vout1 = vec_madd(vprevbot, vkaprev, vout1);

			t1 = vec_mergeh( vout1, vout3 );//1 {a0, c0, a1, c1}
		 	t2 = vec_mergel( vout1, vout3 );//2 {a2, c2, a3, c3}

			vout2 = vec_madd(vcurrtop, vkacurr2, vzero);
			vout4 = vec_madd(vnexttop, vkanext, vzero);
			vout2 = vec_madd(vcurrmid, vkbcurr2, vout2);
			vout4 = vec_madd(vcurrtop, vkacurr4, vout4);
			vout4 = vec_madd(vnextmid, vkbnext, vout4);
			vout4 = vec_madd(vcurrmid, vkbcurr4, vout4);
			vout2 = vec_madd(vcurrbot, vkacurr2, vout2);
			vout4 = vec_madd(vnextbot, vkanext, vout4);
			vout4 = vec_madd(vcurrbot, vkacurr4, vout4);
			
 			t3 = vec_mergeh( vout2, vout4 );//3 {b0, d0, b1, d1}
 			t4 = vec_mergel( vout2, vout4 );//4 {b2, d2, b3, d3}

			//Add intermediate values 
			vout2 = vec_add( t1, t2 ); //4 {a0 + a2, c0 + c2, a1 + a3, c1 + c3}
			vout4 = vec_add( t3, t4 ); //6 {b0 + b2, d0 + d2, b1 + b3, d1 + d3}

			//Do half of the second half of the transpose 
			vout1 = vec_mergeh( vout2, vout4 ); //7 { a0 + a2, b0 + b2, c0 + c2, d0 + d2 }
			vout3 = vec_mergel( vout2, vout4 ); //8 { a1 + a3, b1 + b3, c1 + c3, d1 + d3 }

			// get prev output
			vprevout = vec_ld(j, pr_out2);  

			// subtract previous output
			vout1 = vec_sub(vout1, vprevout);
			
			// get mask
			vmask = vec_ld(j, prm);  	

			// Find the result
			vout1 = vec_add(  vout1,  vout3 ); 
			
			// clip
			vout1 = vec_min(vec_max(vout1, vmin), vmax);
			
			// mult mask
			vout1 = vec_madd(vout1, vmask, vzero);
			
			// store vout
			vec_st(vout1, j, pr_out2);   			
		}
	}


	
	// calculate new border junction values for next step using finite difference relation
	// Pj[t] = (1+r)Padjacent[t-1] - (r)Pj[t-2]	
	{
		float m, mq;
		float a0 = x->bq_a0, a1 = x->bq_a1, a2 = x->bq_a2, b1 = x->bq_b1, b2 = x->bq_b2;
		// correct reflectivity for impedance on power curve from (-1, 1) to (-1, 1) (ad hoc) 
		r = 2. * jit_math_pow(0.5*(x->reflectivity+1), 4.-((3.5)*x->tension)) - 1.;
		
		// top
		pr_out2 = (float *)((long)p_out + (long)((1)*x->rowbytes) ); // t
		pr_in2 = (float *)((long)p_in  + (long)((1)*x->rowbytes)); // t-1
		pr_border = (float *)((long)x->border_matrix_data + (long)((1)*x->rowbytes)); 	// t
		pr_border1 = (float *)((long)x->border1_matrix_data + (long)((1)*x->rowbytes)); // t-1
		pr_border2 = (float *)((long)x->border2_matrix_data + (long)((1)*x->rowbytes)); // t-2
		for(j=1; j<width-1; j++)
		{
			m = pr_out2[j] + e*pr_border[j];		// add prev border to kernel calc edges
			pr_border2[j] = pr_border1[j];				
			pr_border1[j] = pr_border[j];				
			pr_border[j] = ((1. + r)*pr_in2[j] - r*pr_border2[j]);				// calc reflection
			mq = (m * tsq);										// scale cubic waveshape
			pr_out2[j] = m - mq*mq*mq*one_over_tsq*x->cubic;
		}		
		// bottom
		pr_out2 = (float *)((long)p_out + (long)((height-2)*x->rowbytes)); // t
		pr_in2 = (float *)((long)p_in + (long)((height-2)*x->rowbytes)); // t -1
		pr_border = (float *)((long)x->border_matrix_data+ (long)((height-2)*x->rowbytes)); // t
		pr_border1 = (float *)((long)x->border1_matrix_data + (long)((height-2)*x->rowbytes)); // t-1
		pr_border2 = (float *)((long)x->border2_matrix_data+ (long)((height-2)*x->rowbytes)); // t-2
		for(j=1; j<width-1; j++)
		{
			m = pr_out2[j] + e*pr_border[j];		// add prev border to kernel calc edges
			pr_border2[j] = pr_border1[j];				
			pr_border1[j] = pr_border[j];				
			pr_border[j] = ((1. + r)*pr_in2[j] - r*pr_border2[j]);				// calc reflection
			mq = (m * tsq);										// scale cubic waveshape
			pr_out2[j] = m - mq*mq*mq*one_over_tsq*x->cubic;
		}	
		// left, right
		for(i=2; i<height-2; i++)
		{
			pr_out2 = (float *)((long)p_out + (long)((i)*x->rowbytes)); // t
			pr_in2 = (float *)((long)p_in + (long)((i)*x->rowbytes)); // t -1
			pr_border = (float *)((long)x->border_matrix_data+ (long)((i)*x->rowbytes)); // t
			pr_border1 = (float *)((long)x->border1_matrix_data+ (long)((i)*x->rowbytes)); // t-1
			pr_border2 = (float *)((long)x->border2_matrix_data+ (long)((i)*x->rowbytes)); // t-2

			j = 1;
			m = pr_out2[j] + e*pr_border[j];		// add prev border to kernel calc edges
			pr_border2[j] = pr_border1[j];				
			pr_border1[j] = pr_border[j];				
			pr_border[j] = ((1. + r)*pr_in2[j] - r*pr_border2[j]);				// calc reflection
			mq = (m * tsq);										// scale cubic waveshape
			pr_out2[j] = m - mq*mq*mq*one_over_tsq*x->cubic;
				
			j = width-2;
			m = pr_out2[j] + e*pr_border[j];		// add prev border to kernel calc edges
			pr_border2[j] = pr_border1[j];				
			pr_border1[j] = pr_border[j];				
			pr_border[j] = ((1. + r)*pr_in2[j] - r*pr_border2[j]);				// calc reflection
			mq = (m * tsq);										// scale cubic waveshape
			pr_out2[j] = m - mq*mq*mq*one_over_tsq*x->cubic;
		}
	}
	
//   RestoreJavaMode ( &oldJavaMode );
}
*/


inline void run_mesh_step_scalar(t_mesh *x, float *p_in, float *p_out, float *p_mask, float *p_kernel)
{
	long i, j, nk;
	long width = x->dim[0];
	long height = x->dim[1];
	float * pr1, *pr2, *pr3, *pr_border;
	float r = x->reflectivity;
	float tsq = x->tension;
	float one_over_tsq = 1.0 / tsq;
	float * prm;
	float * pr_out;
	float * p_k = p_kernel;
	float f, ff;
	float k, e, c, t;
	long half = width << 1;
	int offset = 0;
	float a0 = x->bq_a0, a1 = x->bq_a1, a2 = x->bq_a2, b1 = x->bq_b1, b2 = x->bq_b2;


	// tiny bit of noise added to prevent denormals
	static const float denorm_noise[3] = {1e-10, 0, -1e-10};
	
	k = p_k[0];	// corner
	e = p_k[1];	// edge
	c = p_k[4];	// center

	nk = x->noisectr;

	// copy previous time step border junction values to border temp matrix. 
	for(i=0; i<height; i++)
	{
		pr_out = (float *)((long)p_out + (long)(i*x->rowbytes));
		pr_border = (float *)((long)x->border_matrix_data + (long)(i*x->rowbytes));
		for(j=0; j<width; j++)
		{
			if ((i==0)||(i==height-1)||(j==0)||(j==width-1))
			{
				pr_border[j] = pr_out[j];
			}
		}	
	}
	
	// run mesh center
	for(i=1; i<height-1; i++)
	{
		// row ptrs
		pr1 = (float *)((long)p_in + (long)((i-1)*x->rowbytes));
		pr2 = (float *)((long)p_in + (long)(i*x->rowbytes));
		pr3 = (float *)((long)p_in + (long)((i+1)*x->rowbytes));
		prm = (float *)((long)p_mask + (long)(i*x->rowbytes));
		pr_out = (float *)((long)p_out + (long)(i*x->rowbytes));	
			
		for(j=1; j<width-1; j++)
		{
			ff = pr_out[j];	
			f = e*(pr2[j-1] + pr1[j] + pr2[j+1] + pr3[j]);
			f += k*(pr1[j-1] + pr1[j+1] + pr3[j-1] + pr3[j+1]);
			f += c*pr2[j];
			f -= ff;
			pr_out[j] = f * prm[j] + denorm_noise[nk++];
			if (nk >= 3) nk=0;
		}
	}
		
	x->noisectr = nk;

	// calculate new border junction values for next step using finite difference relation
	// Pj[t] = (1+r)Padjacent[t-1] - (r)Pj[t-2]	
	// top
	pr_out = (float *)((long)p_out); // t
	pr3 = (float *)((long)p_in + (long)(x->rowbytes)); // t-1
	pr_border = (float *)((long)x->border_matrix_data); // t-2
	for(j=1; j<width-1; j++)
	{
		pr_out[j] = (1. + r)*pr3[j] - r*pr_border[j];		
//		pr_out[j] = t - t*t*t*x->cubic;
	}		
	// bottom
	pr_out = (float *)((long)p_out + (long)((height-1)*x->rowbytes)); // t
	pr1 = (float *)((long)p_in + (long)((height-2)*x->rowbytes)); // t-1
	pr_border = (float *)((long)x->border_matrix_data+ (long)((height-1)*x->rowbytes)); // t-2
	for(j=1; j<width-1; j++)
	{
		pr_out[j] = (1. + r)*pr1[j] - r*pr_border[j];
//		pr_out[j] = t - t*t*t*x->cubic;
	}	
	// left, right
	for(i=1; i<height-1; i++)
	{
		pr_out = (float *)((long)p_out + (long)(i*x->rowbytes)); // t
		pr2 = (float *)((long)p_in + (long)(i*x->rowbytes)); // t-1
		pr_border = (float *)((long)x->border_matrix_data + (long)(i*x->rowbytes)); // t-2
		pr_out[0] = (1. + r)*pr2[1] - r*pr_border[0];
//		pr_out[0] = t - t*t*t*x->cubic;
		pr_out[width-1] = (1. + r)*pr2[width-2] - r*pr_border[width-1];
//		pr_out[width-1] = t - t*t*t*x->cubic;
	}
}



/*
void TurnJavaModeOff( vector unsigned long *oldJavaMode )
{
    vector unsigned long javaOffMask = ( vector unsigned long )( 0x00010000 );
    vector unsigned long java;
    *oldJavaMode = ( vector unsigned long ) vec_mfvscr ( );
    java = vec_or ( *oldJavaMode, javaOffMask );
    vec_mtvscr ( java );
}
      
void RestoreJavaMode( vector unsigned long *oldJavaMode )
{
	vec_mtvscr ( *oldJavaMode );
}
*/



void make_frac_kernel(t_mesh *x)
{
	float t2, k, e, c;	// tension^2, korner, edge, center
	
	t2 = x->tension * x->tension;
	double b0;
	double alpha;
	double cs, w, Q, cutoff;
	
	// equal energy criterion: 4k + 4e + c = 2.0.
	// the simulation is valid up to t^2 = 3/5, at which c=0
	// and the speed of wave travel is one mesh unit per time step.
	//
	// we still get reasonable sounds past this limit, by distributing
	// more of the energy to the corners, until the
	// stability breaks down at c < -0.5.
	
	if (x->interpolate)
	{	
		k = t2 * (1. / 6.) ;
		e = t2 * (2. / 3.) ;
		c = 2. - 4.*(k + e);
	}
	else
	{
		k = 0. ;
		e = t2 * 0.5 ;
		c = 2. - 4.*(k + e);
	}
		
// post ("corner:%f, edge:%f, center:%f", k, e, c);
	
	x->frac_kernel[0] = k;
	x->frac_kernel[1] = e;
	x->frac_kernel[2] = k;
	x->frac_kernel[3] = e;
	x->frac_kernel[4] = c;
	x->frac_kernel[5] = e;
	x->frac_kernel[6] = k;
	x->frac_kernel[7] = e;
	x->frac_kernel[8] = k;
	
	// change filter to match lower "sampling rate" (tension) of interpolated kernel


//	cutoff = x->tension * TENSION_TO_FREQ;  // from measurements

	cutoff = 100.;
	
	Q = 0.66;		// gain never above 1.0 at any frequency
  	
	w = cutoff * (TWOPI / x->fs);
	alpha = sin(w)/(2.*Q);
	cs = cos(w); 
	b0 			= 1. / ( 1. +  alpha);
	
	// also make FIR coefficients
	
	/*
	
	b0 			= 1. / ( 1. +  alpha);
	
	f->f_a0 	= 
	f->f_a2 	= b0 * ((1. - cs) * 0.5);
    f->f_a1 	= b0 * (1.  -  cs);
    f->f_b1 	= b0 * (-2. *  cs);  
    f->f_b2 	= b0 * (1.  -  alpha);
	*/
	
	
	x->bq_a0 = b0 * ((1. - cs) * 0.5);
	x->bq_a1 = b0 * (1.  -  cs);
	x->bq_a2 = b0 * ((1. - cs) * 0.5);
	x->bq_b1 = b0 * (-2. *  cs);
	x->bq_b2 = b0 * (1.  -  alpha);
	
}



// multiply input by mask
void add_input_to_mesh(t_mesh *x)
{
	float *p_dest, *p_in, *p_out, *prm;
	int i, j;
	float f;
	int width = x->dim[0];
	int height = x->dim[1];

	for (i=0; i<height; i++)
	{
		prm = (float *)((long)x->mask_data + (long)(i*x->rowbytes));
		p_in = (float *)((long)x->in1_matrix_data + (long)(i*x->rowbytes));
		p_out = (float *)((long)x->in2_matrix_data + (long)(i*x->rowbytes));
		for(j=0; j<width; j++)
		{
			p_out[j] = p_in[j]*prm[j];
		}
	}
}


void setmem(void *ss, long n, short b)
{
	register int i;
	unsigned char *s = (unsigned char *)ss;
	
	for (i=0; i < n; i++,s++)
		*s = (unsigned char)b;
}
